/*
    lircmix - audio mixer

    This version Copyright (C) 2001 Alastair M. Robinson
    loosely based on smix, which is (C) 1997 by Sven Anders

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <errno.h>
#include <malloc.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <pwd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#ifdef HAVE_LIBASOUND
#include <alsa/asoundlib.h>
#else
#include <linux/soundcard.h>
#endif
#include "mix.h"

#define TRUE 1
#define FALSE 0

/*--------------------------------------------------------------------------*/

#ifndef HAVE_LIBASOUND
char *dev_labels[]=SOUND_DEVICE_LABELS;
#endif

static void rec_channel(struct mixer *mix,int channel);
static void solo_channel(struct mixer *mix,int channel);
static void mute_channel(struct mixer *mix,int channel);
static void unmute_channel(struct mixer *mix,int channel);
static void togglemute_channel(struct mixer *mix,int channel);
static int get_mute_channel(struct mixer *mix, int channel);
static void mixer_inc_volume(struct mixer *mix,int channel,
                             int left, int right);

static void mixer_dec_volume(struct mixer *mix,int channel,
                             int left, int right);

static void mixer_get_volume(struct mixer *mix,int channel,
                             int *left, int *right);

static void mixer_set_volume(struct mixer *mix,int channel,
                             int left, int right);

static void mixer_dispose(struct mixer *mix);
static void print_value(struct mixer *mix,int i);
static char *get_channelname(struct mixer *mix,int channel);
static int get_channeldev(struct mixer *mix,int channel);

/*--------------------------------------------------------------------------*/

#define MIXFAIL(msg) { fprintf(stderr,msg); return(NULL); }

struct mixer *mixer_create(char *dev_name)
{
  struct mixer *mix;
#ifdef HAVE_LIBASOUND
  int i, err;
  snd_ctl_t *ctl_handle;
  snd_mixer_elem_t *elem;
#else
  unsigned int i,j;
#endif
  if (!(mix = calloc(1, sizeof(struct mixer)))) return(NULL);
  
  mix->dispose=mixer_dispose;
  mix->setvolume=mixer_set_volume;
  mix->incvolume=mixer_inc_volume;
  mix->decvolume=mixer_dec_volume;
  mix->getvolume=mixer_get_volume;
  mix->getchannelname=get_channelname;
  mix->getchanneldev=get_channeldev;
  mix->mute=mute_channel;
  mix->unmute=unmute_channel;
  mix->togglemute=togglemute_channel;
  mix->getmute=get_mute_channel;
  mix->solo=solo_channel;
  mix->record=rec_channel;

#ifdef HAVE_LIBASOUND
  snd_ctl_card_info_alloca(&(mix->hw_info));
  err = snd_ctl_open(&ctl_handle, dev_name, 0);
  if (err != 0) MIXFAIL("Can't open control device\n");
  err = snd_ctl_card_info(ctl_handle, mix->hw_info);
  if (err != 0) MIXFAIL("Can't get mixer info\n");
  snd_ctl_close(ctl_handle);
  err = snd_mixer_open(&(mix->handle), 0);
  if (err != 0) MIXFAIL("Can't open mixer device\n");
  err = snd_mixer_attach(mix->handle, dev_name);
  if (err != 0) MIXFAIL("Can't attach to mixer\n");
  err = snd_mixer_selem_register(mix->handle, NULL, NULL);
  if (err != 0) MIXFAIL("Can't config mixer registers\n");
  err = snd_mixer_load(mix->handle);
  if (err != 0) MIXFAIL("Can't load mixer\n");
  mix->devs = 0;
  mix->rec_devs = 0;
  mix->stereo_devs = 0;
  for (elem = snd_mixer_first_elem(mix->handle), i = 0; elem; 
       elem = snd_mixer_elem_next(elem)) 
  {
    if (snd_mixer_selem_is_active(elem) && 
	(snd_mixer_selem_has_playback_volume(elem) || 
	 snd_mixer_selem_has_capture_volume(elem))) 
    {
      mix->channel[i].label = strdup(snd_mixer_selem_get_name(elem));
      mix->channel[i].elem = elem;
      if (snd_mixer_selem_is_playback_mono(elem)) 
      {
	mix->channel[i].stereo = FALSE;
      }
      else 
      {
	mix->channel[i].stereo = TRUE;
	mix->stereo_devs++;
      }
      if (snd_mixer_selem_has_capture_switch(elem)) 
      {
	mix->channel[i].rec_support = TRUE;
	mix->rec_devs++;
	mix->channel[i].rec_src = TRUE;
      } 
      else 
      {
	mix->channel[i].stereo = FALSE;
	mix->channel[i].rec_support = FALSE;
      }
      i++;
    }
    mix->devs = i;
  }
#else
  if ((mix->handle=open(dev_name, O_RDWR))<0)
    MIXFAIL("Can't open device\n");
  if (ioctl(mix->handle, SOUND_MIXER_INFO, &mix->info)==-1)
    MIXFAIL("Can't get mixer info\n");
  if (ioctl(mix->handle, SOUND_MIXER_READ_DEVMASK, &mix->devmask)==-1)
    MIXFAIL("Can't get devmask\n");
  if (ioctl(mix->handle, SOUND_MIXER_READ_STEREODEVS, &mix->stereomask)==-1)
    MIXFAIL("Can't get stereodevs\n");
  if (ioctl(mix->handle, SOUND_MIXER_READ_RECMASK, &mix->recmask)==-1)
    MIXFAIL("Can't get recmask\n");
  if (ioctl(mix->handle, SOUND_MIXER_READ_RECSRC, &mix->recsrc)==-1)
    MIXFAIL("Can't get recsrc\n");

  if (!mix->devmask)
      MIXFAIL("No mixer devices to adjust!\n");
  mix->devs = 0;
  mix->rec_devs = 0;
  mix->stereo_devs = 0;

  for (i=0, j=1; i<MAX_NUM_MIXER_DEVS; i++, j<<=1)
  {
    if (mix->devmask & j)
    {
      mix->channel[mix->devs].label = (char *)malloc(strlen(dev_labels[i])+1);
      strcpy(mix->channel[mix->devs].label, dev_labels[i]);

      if ((mix->stereomask & j) != 0)
      {
        mix->channel[mix->devs].stereo = TRUE;
        mix->stereo_devs++;
      }
      else
        mix->channel[mix->devs].stereo = FALSE;

      if ((mix->recmask & j) != 0)
      {
        mix->channel[mix->devs].rec_support = TRUE;
        mix->rec_devs++;
      }
      else
        mix->channel[mix->devs].rec_support = FALSE;

      if ((mix->recsrc & j) != 0)
        mix->channel[mix->devs].rec_src = TRUE;
      else
        mix->channel[mix->devs].rec_src = FALSE;

      mix->channel[mix->devs].dev_num = i;
      mix->channel[mix->devs].dev_bit = j;
      mix->getvolume(mix,mix->devs, &mix->channel[mix->devs].left,
                                    &mix->channel[mix->devs].right);

      mix->channel[mix->devs].muted_l = -1;
      mix->channel[mix->devs].muted_r = -1;

      mix->devs++;
    }
  }
#endif

  return(mix);
}

static void mixer_set_volume(struct mixer *mix,int channel,
                             int left, int right)
{
#ifdef HAVE_LIBASOUND
  snd_mixer_elem_t *elem = mix->channel[channel].elem;
  snd_mixer_selem_set_playback_switch_all(elem, 1);
  if (snd_mixer_selem_has_playback_volume(elem)) 
  {
    snd_mixer_selem_set_playback_volume(elem, SND_MIXER_SCHN_FRONT_LEFT, left);
    if (!snd_mixer_selem_is_playback_mono(elem))
      snd_mixer_selem_set_playback_volume(elem, SND_MIXER_SCHN_FRONT_RIGHT, right);
  }
  else 
  {
    snd_mixer_selem_set_capture_volume(elem, SND_MIXER_SCHN_FRONT_LEFT, left);
    if (!snd_mixer_selem_is_playback_mono(elem))
      snd_mixer_selem_set_capture_volume(elem, SND_MIXER_SCHN_FRONT_RIGHT, right);
  }
#else
  unsigned char temp[4];

  mix->channel[channel].left = left;
  mix->channel[channel].right = right;
  temp[0]=(unsigned char)left;
  temp[1]=(unsigned char)right;
  temp[2]=temp[3]=0;
  ioctl(mix->handle, MIXER_WRITE(mix->channel[channel].dev_num), temp);
#endif
}


static void mixer_dec_volume(struct mixer *mix,int channel,
                             int left, int right)
{
  mix->incvolume(mix,channel,-left,-right);
}


static void mixer_inc_volume(struct mixer *mix,int channel,
                             int left, int right)
{
  unsigned char temp[4];

#ifdef HAVE_LIBASOUND
  snd_mixer_selem_set_playback_switch_all(mix->channel[channel].elem, 1);
#endif
  mix->getvolume(mix,channel,&mix->channel[channel].left,
                             &mix->channel[channel].right);
  left+=mix->channel[channel].left;
  right+=mix->channel[channel].right;

  if(left<0) left=0;
  if(right<0) right=0;
  if(left>MAX_MIXER_VALUE) left=MAX_MIXER_VALUE;
  if(right>MAX_MIXER_VALUE) right=MAX_MIXER_VALUE;
  
  mix->channel[channel].left = left;
  mix->channel[channel].right = right;

#ifdef HAVE_LIBASOUND
  mixer_set_volume(mix, channel, left, right);
#else
  temp[0]=(unsigned char)left;
  temp[1]=(unsigned char)right;
  temp[2]=temp[3]=0;
  ioctl(mix->handle, MIXER_WRITE(mix->channel[channel].dev_num), temp);
#endif
}


static void mixer_get_volume(struct mixer *mix,int channel,
                             int *left, int *right)
{
#ifdef HAVE_LIBASOUND
  long left_chn, right_chn, pmin, pmax;
  snd_mixer_elem_t *elem = mix->channel[channel].elem;  
  if (snd_mixer_selem_has_playback_volume(elem))
    snd_mixer_selem_get_playback_volume_range(elem, &pmin, &pmax);
  else
    snd_mixer_selem_get_capture_volume_range(elem, &pmin, &pmax);
  
  if (snd_mixer_selem_has_playback_volume(elem))
    snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_FRONT_LEFT, &left_chn);
  else
    snd_mixer_selem_get_capture_volume(elem, SND_MIXER_SCHN_FRONT_LEFT, &left_chn);
    
  if (snd_mixer_selem_is_playback_mono(elem))
  {
    left_chn = right_chn = (int) (100 - (*left * (100 / (float) pmax)));
  }
  else
  {
    if (snd_mixer_selem_has_playback_volume(elem))
      snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_FRONT_RIGHT, &right_chn);
    else
      snd_mixer_selem_get_capture_volume(elem, SND_MIXER_SCHN_FRONT_RIGHT, &right_chn);
  }
  *left = (int) left_chn;
  *right = (int) right_chn;
#else
  unsigned char temp[4];

  ioctl(mix->handle, MIXER_READ(mix->channel[channel].dev_num), &temp);
  (unsigned char) *left = temp[0];
  (unsigned char) *right = temp[1];
#endif
}


static void mixer_dispose(struct mixer *mix)
{
  if(mix)
  {
    int i;
    for(i = 0; i < MAX_NUM_MIXER_DEVS; i++) {
      if (mix->channel[i].label != NULL)
	free(mix->channel[i].label);
    }
    free(mix);
  }
}


static void mute_channel(struct mixer *mix,int channel)
{
#ifdef HAVE_LIBASOUND
    snd_mixer_selem_set_playback_switch_all(mix->channel[channel].elem, 0);
#else
  int templ,tempr;

  if((mix->channel[channel].muted_l==-1)||(mix->channel[channel].muted_r==-1))
  {
    mix->getvolume(mix,channel,&templ,&tempr);
    mix->channel[channel].muted_l = templ;
    mix->channel[channel].left = 0;
    mix->channel[channel].muted_r = tempr;
    mix->channel[channel].right = 0;
    mix->setvolume(mix,channel, mix->channel[channel].left,
                                mix->channel[channel].right);
  }
#endif
}


static void unmute_channel(struct mixer *mix,int channel)
{
#ifdef HAVE_LIBASOUND
  snd_mixer_selem_set_playback_switch_all(mix->channel[channel].elem, 1);
#else
  if((mix->channel[channel].muted_l>-1)||(mix->channel[channel].muted_r>-1))
  {
    mix->channel[channel].left = mix->channel[channel].muted_l;
    mix->channel[channel].muted_l = -1;
    mix->channel[channel].right = mix->channel[channel].muted_r;
    mix->channel[channel].muted_r = -1;
    mix->setvolume(mix,channel, mix->channel[channel].left,
                                mix->channel[channel].right);
  }
#endif
}


static void togglemute_channel(struct mixer *mix,int channel)
{
#ifdef HAVE_LIBASOUND
  int val;
  snd_mixer_selem_get_playback_switch(mix->channel[channel].elem, 0, &val);
  snd_mixer_selem_set_playback_switch_all(mix->channel[channel].elem, !val);
#else
  if((mix->channel[channel].muted_l==-1)||(mix->channel[channel].muted_r==-1))
    mix->mute(mix,channel);
  else
    mix->unmute(mix,channel);
#endif
}


static int get_mute_channel(struct mixer *mix, int channel)
{
#ifdef HAVE_LIBASOUND
  int val;
  snd_mixer_selem_get_playback_switch(mix->channel[channel].elem, 0, &val);
  return val;
#else
  return ((mix->channel[channel].muted_l==-1)||(mix->channel[channel].muted_r==-1)) ? 1 : 0;
#endif
}

static void solo_channel(struct mixer *mix,int channel)
{
  int i;

  for (i=0; i<mix->devs; i++)
  {
    if (mix->channel[i].solo_channel)
    {
      if (i == channel)
      {
        unmute_channel(mix,i);
      }
      else
      {
        mute_channel(mix,i);
      }
    }
  }
}


static void rec_channel(struct mixer *mix,int channel)
{
#ifdef HAVE_LIBASOUND
  snd_mixer_selem_set_capture_switch_all(mix->channel[channel].elem, 1);
#else
  int i;

  for (i=0; i<mix->devs; i++)
  {
    if (i == channel)
    {
      mix->channel[channel].rec_src = TRUE;
      mix->recsrc |= mix->channel[channel].dev_bit;
    }
    else
    {
      mix->channel[channel].rec_src = FALSE;
      mix->recsrc &= (0xffffffff ^ mix->channel[i].dev_bit);
    }
  }

  ioctl(mix->handle, SOUND_MIXER_WRITE_RECSRC, &mix->recsrc)==-1;
#endif
}


static char *get_channelname(struct mixer *mix,int channel)
{
  if(channel<mix->devs)
    return(mix->channel[channel].label);
  else
    return(NULL);
}


static int get_channeldev(struct mixer *mix,int channel)
{
  if(channel<mix->devs)
    return(mix->channel[channel].dev_num);
  else
    return(-1);
}
